# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

from __future__ import annotations
import numpy as np
from dataclasses import dataclass, field
from typing import Literal



@dataclass
class Point:
    """ A representation of a point."""
    x: float
    y: float
    z: float

    
@dataclass
class Axis:
    """A representation of an axis.
    An Axis object always has length 1 and points in some 3D direction.
    By default XAX, YAX, and ZAX are constructed and defined in the global namespace of the
    FEM module.
    """
    vector: np.ndarray

    def __repr__(self) -> str:
        return f"Axis({self.vector})"
    
    def __post_init__(self):

        self.vector = self.vector/np.linalg.norm(self.vector)
        self.np: np.ndarray = self.vector
        self.x: float = self.vector[0]
        self.y: float = self.vector[1]
        self.z: float = self.vector[2]

    def __neg__(self):
        return Axis(-self.vector)
    
    @property
    def neg(self) -> Axis:
        return Axis(-self.vector)
    
    def cross(self, other: Axis) -> Axis:
        """Take the cross produt with another vector

        Args:
            other (Axis): Vector B in AxB

        Returns:
            Axis: The resultant Axis.
        """
        return Axis(np.cross(self.vector, other.vector))

    def dot(self, other: Axis) -> float:
        """Take the dot product of two vectors A·B

        Args:
            other (Axis): Vector B

        Returns:
            float: The resultant vector (A·B)
        """
        return np.dot(self.vector, other.vector)
    
    def pair(self, other: Axis) -> Plane:
        """Pair this vector with another to span the plane A⨂B

        Args:
            other (Axis): Vector B

        Returns:
            Plane: The plane spanned by A and B.
        """
        return Plane(self, other)
    
    def __mul__(self, other: Axis) -> Plane:
        """The multiply binary operator

        Args:
            other (Axis): _description_

        Returns:
            Plane: The output plane
        """
        return self.pair(other)
    
    def construct_cs(self) -> CoordinateSystem:
        """Constructs a coordinate system where this vector is the Z-axis
        and the X and Y axis are normal to this axis but with an arbitrary rotation.

        Returns:
            CoordinateSystem: The resultant coordinate system
        """
        ax = Axis(np.array([1, 0, 0]))
        if np.abs(self.dot(ax)) > 0.999:
            
            ax = Axis(np.array([0, 1, 0]))
        ax1 = self.cross(ax)
        ax2 = self.cross(ax1).neg
        return CoordinateSystem(ax2, ax1, self, np.zeros(3))

XAX: Axis = Axis(np.array([1, 0, 0]))
YAX: Axis = Axis(np.array([0, 1, 0]))
ZAX: Axis = Axis(np.array([0, 0, 1]))

def _parse_vector(vec: np.ndarray | tuple[float, float, float] | list[float] | Axis) -> np.ndarray:
    """ Takes an array, tuple, list or Axis and alwasys returns an array."""
    if isinstance(vec, np.ndarray):
        return vec
    elif isinstance(vec, (list,tuple)):
        return np.array(vec)
    elif isinstance(vec, Axis):
        return vec.vector
    return None

def _parse_axis(vec: np.ndarray | tuple[float, float, float] | list[float] | Axis) -> Axis:
    """Takes an array, tuple, list or Axis and always returns an Axis.

    Args:
        vec (np.ndarray | tuple | list | Axis): The Axis data

    Returns:
        Axis: The Axis object.
    """
    if isinstance(vec, np.ndarray):
        return Axis(vec)
    elif isinstance(vec, (list,tuple)):
        return Axis(np.array(vec))
    elif isinstance(vec, Axis):
        return vec
    return None

@dataclass
class Plane:
    """A generalization of any plane of inifinite size spanned by two Axis objects.

    """
    uax: Axis
    vax: Axis

    def __post_init__(self):
        # Check if axes are orthogonal
        if not np.isclose(np.dot(self.uax.vector, self.vax.vector), 0):
            raise ValueError("Axes are not orthogonal")

    def __repr__(self) -> str:
        return f"Plane({self.uax}, {self.vax})"
    
    def __mul__(self, other: Axis) -> CoordinateSystem:
        return CoordinateSystem(self.uax, self.vax, other)
    
    @property
    def normal(self) -> Axis:
        """Returns the normal of the plane as u ⨉ v.

        Returns:
            Axis: The axis object normal to the plane.
        """
        return self.uax.cross(self.vax)
    
    def flip(self) -> Plane:
        """Flips the planes U and V axes.

        Returns:
            Plane: A new plane object.
        """
        return Plane(self.vax, self.uax)
    
    def cs(self, x0: float = 0, y0: float = 0, z0: float = 0) -> CoordinateSystem:
        """Returns a CoordinateSystem object for the plane where the XY axes are aligned
        with the plane UV axis and Z is normal.

        Args:
            x0 (float): The x coordinate of the origin
            y0 (float): The y coordinate of the origin
            z0 (float): The z coordinate of the origin

        Returns:
            CoordinateSystem: The coordinate system object
        """
        origin = np.array([x0, y0, z0])
        return CoordinateSystem(self.uax, self.vax, self.normal, origin)
    
    def grid(self, 
             uax: np.ndarray | tuple[float, float, int], 
             vax: np.ndarray | tuple[float, float, int], 
             origin: np.ndarray | tuple[float, float, float],
             indexing: Literal['xy','ij'] = 'xy') -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Spans a grid of points in the plane based on a np.linspace like argument type.
        The first uax argument should be a start, finish, Npoints type tuple of a float, float and integer.
        Item for the vax. The origin defines the coordinate at which u,v = 0 will be placed.
        The return type is an N,M np.meshgrid defined by the indexing 'xy' or 'ij'.

        Args:
            uax (np.ndarray | tuple[float, float, int]): The uax linspace argument values
            vax (np.ndarray | tuple[float, float, int]): The vax linspace argument values
            origin (np.ndarray | tuple[float, float, float]): The origin for u,v = 0
            indexing (Literal[&#39;xy&#39;,&#39;ij&#39;], optional): The indexing type. Defaults to 'xy'.

        Returns:
            tuple[np.ndarray, np.ndarray, np.ndarray]: The X, Y, Z (N,M) meshgrid of coordinates.
        """
        if isinstance(uax, tuple):
            uax = np.linspace(*uax)
        if isinstance(vax, tuple):
            vax = np.linspace(*vax)
        if isinstance(origin, tuple):
            origin = np.array(origin)

        U, V = np.meshgrid(uax, vax, indexing=indexing)
        uax = U.flatten()
        vax = V.flatten()
        shp = U.shape
        xs = self.uax.np[0]*uax + self.vax.np[0]*vax + origin[0]
        ys = self.uax.np[1]*uax + self.vax.np[1]*vax + origin[1]
        zs = self.uax.np[2]*uax + self.vax.np[2]*vax + origin[2]

        return xs.reshape(shp), ys.reshape(shp), zs.reshape(shp)
    
    def span(self, u: float, v: float, N: int, origin: np.ndarray) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Create a grid of XYZ coordinates in the plane reaching from 0 to u and 0 to v in N steps at the given origin

        Args:
            u (float): The distance along the first axis
            v (float): The distance along the second axis 
            N (int): The number of sample points per axis
            origin (np.ndarray): The origin of the planar grid.

        Returns:
            tuple[np.ndarray, np.ndarray, np.ndarray]: The set of X,Y,Z coordinates.
        """
        uax = np.linspace(0, u, N)
        vax = np.linspace(0, v, N)
        U, V = np.meshgrid(uax, vax, indexing='ij')
        uax = U.flatten()
        vax = V.flatten()
        shp = U.shape
        xs = self.uax.np[0]*uax + self.vax.np[0]*vax + origin[0]
        ys = self.uax.np[1]*uax + self.vax.np[1]*vax + origin[1]
        zs = self.uax.np[2]*uax + self.vax.np[2]*vax + origin[2]

        return xs.reshape(shp), ys.reshape(shp), zs.reshape(shp)

XYPLANE = Plane(XAX, YAX)
XZPLANE = Plane(XAX, ZAX)
YZPLANE = Plane(YAX, ZAX)
YXPLANE = Plane(YAX, XAX)
ZXPLANE = Plane(ZAX, XAX)
ZYPLANE = Plane(ZAX, YAX)


@dataclass
class CoordinateSystem:
    """A class representing CoordinateSystem information.

    This class is widely used throughout the FEM solver to embed objects in space properly.
    The x,y and z unit vectors are defined by Axis objects. The origin by a np.ndarray.

    The property _is_global is should only be set for any CoordinateSystem class that is wished to
    be considered as global. This is reserved for the GCS intance create automatically with:
        xax = (1,0,0)
        yax = (0,1,0)
        zax = (0,0,1)
        origin = (0., 0., 0.) meters
    """
    xax: Axis
    yax: Axis
    zax: Axis
    origin: np.ndarray = field(default_factory=lambda: np.array([0,0,0]))
    _is_global: bool = False

    def __post_init__(self):
        self.xax = _parse_axis(self.xax)
        self.yax = _parse_axis(self.yax)
        self.zax = _parse_axis(self.zax)

        self._basis = np.array([self.xax.np, self.yax.np, self.zax.np]).T

        self._basis_inv = np.linalg.pinv(self._basis)

    def __repr__(self) -> str:
        return f"CoordinateSystem({self.xax}, {self.yax}, {self.zax}, {self.origin})"
    
    def copy(self) -> CoordinateSystem:
        """ Creates a copy of this coordinate system."""
        return CoordinateSystem(self.xax, self.yax, self.zax, self.origin)
    
    def displace(self, dx: float, dy: float, dz: float) -> CoordinateSystem:
        """Creates a displaced version of this coordinate system. The basis is kept the same.

        Args:
            dx (float): The X-displacement (meters)
            dy (float): The Y-displacement (meters)
            dz (float): The Z-displacement (meters)

        Returns:
            CoordinateSystem: The new CoordinateSystem object.
        """
        csnew = CoordinateSystem(self.xax, self.yax, self.zax, self.origin + np.array([dx, dy, dz]))
        return csnew
    
    def rotate(self, axis: tuple | list | np.ndarray | Axis, 
               angle: float, 
               degrees: bool = True) -> CoordinateSystem:
        """Return a new CoordinateSystem rotated about the given axis (through the global origin)
        by `angle`. If `degrees` is True, `angle` is interpreted in degrees.

        Args:
            axis (tuple | list | np.ndarray | Axis): The rotation axis
            angle (float): The rotation angle (in degrees if degrees = True)
            degrees (bool, optional): Whether to use degrees. Defaults to True.

        Returns:
            CoordinateSystem: The new rotated coordinate system
        """

        # Convert to radians if needed
        if degrees:
            theta = angle * np.pi/180

        # Normalize the rotation axis
        u = _parse_vector(axis)
        u = u / np.linalg.norm(u)

        # Build the skew-symmetric cross-product matrix K for u
        K = np.array([
            [   0,   -u[2],  u[1]],
            [ u[2],     0,  -u[0]],
            [-u[1],  u[0],     0]
        ], dtype=float)

        # Rodrigues' rotation formula: R = I + sinθ·K + (1−cosθ)·K²
        Imat = np.eye(3, dtype=float)
        R = Imat + np.sin(theta) * K + (1 - np.cos(theta)) * (K @ K)

        # Rotate each axis and the origin
        new_x = R @ self.xax.vector
        new_y = R @ self.yax.vector
        new_z = R @ self.zax.vector
        #new_o = R @ self.origin

        return CoordinateSystem(
            xax=new_x,
            yax=new_y,
            zax=new_z,
            origin=self.origin,
            _is_global=self._is_global
        )
    
    def swapxy(self) -> None:
        """Swaps the XY axes of the CoordinateSystem.
        """
        self.xax, self.yax = self.yax, self.xax
        self.__post_init__()
    
    def affine_from_global(self) -> np.ndarray:
        """Returns an Affine transformation matrix in order to transform coordinates from
        the global coordinate system to this coordinate system.

        Returns:
            np.ndarray: The affine transformation matrix.
        """
        # ensure they’re 1-D
        x = self.xax.vector
        y = self.yax.vector
        z = self.zax.vector
        o = self.origin

        T = np.eye(4, dtype=float)
        T[0:3, 0] = x
        T[0:3, 1] = y
        T[0:3, 2] = z
        T[0:3, 3] = o
        return T
    
    def affine_to_global(self) -> np.ndarray:
        """Returns an Affine transformation matrix in order to transform coordinates from
        this local coordinate system to the coordinate system.

        Returns:
            np.ndarray: The affine transformation matrix.
        """
        T = self.affine_from_global()
        R = T[0:3, 0:3]
        o = T[0:3, 3]
        R_inv = np.linalg.inv(R)
        o_new = - R_inv @ o
        T_inv = np.eye(4, dtype=float)
        T_inv[0:3, 0:3] = R_inv
        T_inv[0:3, 3]   = o_new
        return T_inv
    
    def in_global_cs(self, x: np.ndarray, y: np.ndarray, z: np.ndarray) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Converts x,y,z coordinates into the global coordinate system.

        Args:
            x (np.ndarray): The x-coordinates (meter)
            y (np.ndarray): The y-coordinates (meter)
            z (np.ndarray): The z-coordinates (meter)

        Returns:
            tuple[np.ndarray, np.ndarray, np.ndarray]: The resultant x, y and z coordinates.
        """
        xg = self.xax.np[0]*x + self.yax.np[0]*y + self.zax.np[0]*z + self.origin[0]
        yg = self.xax.np[1]*x + self.yax.np[1]*y + self.zax.np[1]*z + self.origin[1]
        zg = self.xax.np[2]*x + self.yax.np[2]*y + self.zax.np[2]*z + self.origin[2]
        return xg, yg, zg
    
    def in_local_cs(self, x: np.ndarray,
                    y: np.ndarray,
                    z: np.ndarray) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Converts x,y,z coordinates into the local coordinate system.

        Args:
            x (np.ndarray): The x-coordinates (meter)
            y (np.ndarray): The y-coordinates (meter)
            z (np.ndarray): The z-coordinates (meter)

        Returns:
            tuple[np.ndarray, np.ndarray, np.ndarray]: The resultant x, y and z coordinates.
        """
        B = self._basis_inv
        xg = x - self.origin[0]
        yg = y - self.origin[1]
        zg = z - self.origin[2]
        x = B[0,0]*xg + B[0,1]*yg + B[0,2]*zg
        y = B[1,0]*xg + B[1,1]*yg + B[1,2]*zg
        z = B[2,0]*xg + B[2,1]*yg + B[2,2]*zg
        return x, y, z
    
    def in_global_basis(self, x: np.ndarray,
                        y: np.ndarray,
                        z: np.ndarray) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Converts x,y,z vector components into the global coordinate basis.

        Args:
            x (np.ndarray): The x-vector components (meter)
            y (np.ndarray): The y-vector components (meter)
            z (np.ndarray): The z-vector components (meter)

        Returns:
            tuple[np.ndarray, np.ndarray, np.ndarray]: The resultant x, y and z vectors.
        """
        xg = self.xax.np[0]*x + self.yax.np[0]*y + self.zax.np[0]*z
        yg = self.xax.np[1]*x + self.yax.np[1]*y + self.zax.np[1]*z
        zg = self.xax.np[2]*x + self.yax.np[2]*y + self.zax.np[2]*z
        return xg, yg, zg
    
    def in_local_basis(self, x: np.ndarray,
                          y: np.ndarray,
                          z: np.ndarray) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Converts x,y,z vector components into the local coordinate basis.

        Args:
            x (np.ndarray): The x-vector components (meter)
            y (np.ndarray): The y-vector components (meter)
            z (np.ndarray): The z-vector components (meter)

        Returns:
            tuple[np.ndarray, np.ndarray, np.ndarray]: The resultant x, y and z vectors.
        """
        B = self._basis_inv
        xg = x
        yg = y
        zg = z
        x = B[0,0]*xg + B[0,1]*yg + B[0,2]*zg
        y = B[1,0]*xg + B[1,1]*yg + B[1,2]*zg
        z = B[2,0]*xg + B[2,1]*yg + B[2,2]*zg
        return x, y, z
    
    @property
    def gx(self) -> float:
        """ The origin x-coordinate in global coordinates."""
        return self.origin[0]
    
    @property
    def gy(self) -> float:
        """ The origin y-coordinate in global coordinates."""
        return self.origin[1]
    
    @property
    def gz(self) -> float:
        """ The origin z-coordinate in global coordinates."""
        return self.origin[2]
    
    @property
    def gxhat(self) -> np.ndarray:
        """ The x-axis unit vector in global coordinates."""
        return self.xax.np
    
    @property
    def gyhat(self) -> np.ndarray:
        """ The y-axis unit vector in global coordinates."""
        return self.yax.np
    
    @property
    def gzhat(self) -> np.ndarray:
        """ The z-axis unit vector in global coordinates."""
        return self.zax.np

# A shorthand alias for the CoordinateSystem Class 
CS = CoordinateSystem

# The global coordinate system
GCS = CoordinateSystem(XAX, YAX, ZAX, np.zeros(3), _is_global=True)




# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

from __future__ import annotations

from ..cs import CoordinateSystem, GCS, Axis
from ..geometry import GeoPolygon, GeoVolume, GeoSurface
from ..material import Material, AIR, COPPER
from .shapes import Box, Plate, Cyllinder
from .polybased import XYPolygon
from .operations import change_coordinate_system
from .pcb_tools.macro import parse_macro
from .pcb_tools.calculator import PCBCalculator

import numpy as np
from loguru import logger
from typing import Literal, Callable, overload
from dataclasses import dataclass

import math
import gmsh

############################################################
#                        EXCEPTIONS                        #
############################################################

class RouteException(Exception):
    pass

############################################################
#                         CONSTANTS                        #
############################################################

SIZE_NAMES = Literal['0402','0603','1005','1608','2012','3216','3225','4532','5025','6332']
_SMD_SIZE_DICT = {x: (float(x[:2])*0.05, float(x[2:])*0.1) for x in ['0402','0603','1005','1608','2012','3216','3225','4532','5025','6332']}


############################################################
#                         FUNCTIONS                        #
############################################################


def approx(a,b):
    return abs(a-b) < 1e-8

def normalize(vector: np.ndarray) -> np.ndarray:
    norm = np.linalg.norm(vector)
    if norm == 0:
        return vector
    return vector / norm

def _rot_mat(angle: float) -> np.ndarray:
    """Returns a 2D rotation matrix given an angle in degrees

    Args:
        angle (float): The angle in degrees

    Returns:
        np.ndarray: The rotation matrix
    """
    ang = -angle * np.pi/180
    return np.array([[np.cos(ang), -np.sin(ang)], [np.sin(ang), np.cos(ang)]])

############################################################
#                          CLASSES                         #
############################################################


class PCBPoly:

    def __init__(self, 
                 xs: list[float],
                 ys: list[float],
                 z: float = 0,
                 material: Material = COPPER):
        self.xs: list[float] = xs
        self.ys: list[float] = ys
        self.z: float = z
        self.material: Material = material

    @property
    def xys(self) -> list[tuple[float, float]]:
        return list([(x,y) for x,y in zip(self.xs, self.ys)])

@dataclass
class Via:
    x: float
    y: float
    z1: float
    z2: float
    radius: float
    segments: int

class RouteElement:

    def __init__(self):
        self.width: float = None
        self.x: float = None
        self.y: float = None
        self.direction: np.ndarray = None
        self.dirright: np.ndarray = None
    
    @property
    def xy(self) -> tuple[float, float]:
        return self.x, self.y

    @property
    def nr(self) -> tuple[float,float]:
        return (self.x + self.dirright[0]*self.width/2, self.y + self.dirright[1]*self.width/2)
    
    @property
    def nl(self) -> tuple[float,float]:
        return (self.x - self.dirright[0]*self.width/2, self.y - self.dirright[1]*self.width/2)
    
    @property
    def right(self) -> list[tuple[float, float]]:
        raise NotImplementedError()

    @property
    def left(self) -> list[tuple[float, float]]:
        raise NotImplementedError()
    
    def __eq__(self, other: RouteElement) -> bool:
        return approx(self.x, other.x) and approx(self.y, other.y) and (1-abs(np.sum(self.direction*other.direction)))<1e-8
    
class StripLine(RouteElement):

    def __init__(self,
                 x: float,
                 y: float,
                 width: float,
                 direction: tuple[float, float]):
        self.x = x
        self.y = y
        self.width = width
        self.direction = normalize(np.array(direction))
        self.dirright = np.array([self.direction[1], -self.direction[0]])

    def __str__(self) -> str:
        return f'StripLine[{self.x},{self.y},w={self.width},d=({self.direction})]'
    @property
    def right(self) -> list[tuple[float, float]]:
        return [(self.x + self.width/2 * self.dirright[0], self.y + self.width/2 * self.dirright[1])]

    @property
    def left(self) -> list[tuple[float, float]]:
        return [(self.x - self.width/2 * self.dirright[0], self.y - self.width/2 * self.dirright[1])]
      
class StripTurn(RouteElement):

    def __init__(self,
                 x: float,
                 y: float,
                 width: float,
                 direction: tuple[float, float],
                 angle: float,
                 corner_type: str = 'round',
                 champher_distance: float | None = None):
        self.xold: float = x
        self.yold: float = y
        self.width: float = width
        self.old_direction: np.ndarray = normalize(np.array(direction))
        self.direction: np.ndarray = _rot_mat(angle) @ self.old_direction
        self.angle: float = angle
        self.corner_type: str = corner_type
        self.dirright: np.ndarray = np.array([self.old_direction[1], -self.old_direction[0]])
        
        if champher_distance is None:
            self.champher_distance: float = 0.75 * self.width*np.tan(np.abs(angle)/2*np.pi/180)
        else:
            self.champher_distance: float = champher_distance

        turnvec = _rot_mat(angle) @ self.dirright * self.width/2

        if angle > 0:
            self.x = x + width/2 * self.dirright[0] - turnvec[0]
            self.y = y + width/2 * self.dirright[1] - turnvec[1]
        else:
            self.x = x - width/2 * self.dirright[0] + turnvec[0]
            self.y = y - width/2 * self.dirright[1] + turnvec[1]

    def __str__(self) -> str:
        return f'StripTurn[{self.x},{self.y},w={self.width},d=({self.direction})]'
    
    @property
    def right(self) -> list[tuple[float, float]]:
        if self.angle > 0:
            return []
        
        #turning left
        xl = self.xold - self.width/2 * self.dirright[0]
        yl = self.yold - self.width/2 * self.dirright[1]
        xr = self.xold + self.width/2 * self.dirright[0]
        yr = self.yold + self.width/2 * self.dirright[1]

        dist = min(self.width*np.sqrt(2), self.width * np.tan(np.abs(self.angle)/2*np.pi/180))        

        dcorner = self.width*(_rot_mat(self.angle) @ self.dirright)

        xend = xl + dcorner[0]
        yend = yl + dcorner[1]

        out = [(xend, yend)]

        if self.corner_type == 'champher':
            dist = max(0.0, dist - self.champher_distance)
        
        if dist==0:
            return  out
        
        x1 = xr + dist * self.old_direction[0]
        y1 = yr + dist * self.old_direction[1]

        if self.corner_type == 'square':
            return [(x1, y1), (xend, yend)]
        if self.corner_type == 'champher':
            x2 = xend - dist * self.direction[0]
            y2 = yend - dist * self.direction[1]

            return [(x1, y1), (x2, y2), (xend, yend)]
        else:
            raise RouteException(f'Trying to route a StripTurn with an unknown corner type: {self.corner_type}')
    @property
    def left(self) -> list[tuple[float, float]]:
        if self.angle < 0:
            return []
        
        #turning right
        xl = self.xold - self.width/2 * self.dirright[0]
        yl = self.yold - self.width/2 * self.dirright[1]
        xr = self.xold + self.width/2 * self.dirright[0]
        yr = self.yold + self.width/2 * self.dirright[1]
        
        dist = min(self.width*np.sqrt(2), self.width * np.tan(np.abs(self.angle)/2*np.pi/180))        

        dcorner = self.width*(_rot_mat(self.angle) @ -self.dirright)

        xend = xr + dcorner[0]
        yend = yr + dcorner[1]

        out = [(xend, yend)]

        if self.corner_type == 'champher':
            dist =max(0.0, dist - self.champher_distance)
        
        if dist==0:
            return  out
        
        x1 = xl + dist * self.old_direction[0]
        y1 = yl + dist * self.old_direction[1]

        if self.corner_type == 'square':
            return [(xend, yend), (x1, y1)]
        if self.corner_type == 'champher':
            x2 = xend - dist * self.direction[0]
            y2 = yend - dist * self.direction[1]

            return [(xend, yend), (x2, y2), (x1, y1)]
        else:
            raise RouteException(f'Trying to route a StripTurn with an unknown corner type: {self.corner_type}')
 
class StripCurve(StripTurn):
    def __init__(self,
                 x: float,
                 y: float,
                 width: float,
                 direction: tuple[float, float],
                 angle: float,
                 radius: float,
                 dang: float = 10.0):
        self.xold: float = x
        self.yold: float = y
        self.width: float = width
        self.old_direction: np.ndarray = normalize(np.array(direction))
        self.direction: np.ndarray = _rot_mat(angle) @ self.old_direction
        self.angle: float = angle
        self.radius: float = radius
        self.dirright: np.ndarray = np.array([self.old_direction[1], -self.old_direction[0]])
        self.dang: float = dang

        angd = abs(angle*np.pi/180)
        self.start = np.array([x,y])
        self.circ_origin = self.start + radius * np.sign(angle) * self.dirright

        self._xhat = -self.dirright * np.sign(angle)
        self._yhat = self.old_direction
        
        self.end = self.circ_origin + radius*(self._xhat*np.cos(angd)+self._yhat*np.sin(angd))
        self.x, self.y = self.end

    def __str__(self) -> str:
        return f'StripCurve[{self.x},{self.y},w={self.width},d=({self.direction})]'

    @property
    def right(self) -> list[tuple[float, float]]:
        points: list[tuple[float, float]] = []
        Npts = int(np.ceil(abs(self.angle/self.dang)))
        R = self.radius-np.sign(self.angle)*self.width/2
        print(R, self.circ_origin, self._xhat, self._yhat)
        for i in range(Npts):
            ang = abs((i+1)/Npts * self.angle * np.pi/180)
            pnew = self.circ_origin + R*(self._xhat*np.cos(ang)+self._yhat*np.sin(ang))
            points.append((pnew[0], pnew[1]))
        
        return points

    @property
    def left(self) -> list[tuple[float, float]]:
        points: list[tuple[float, float]] = []

        Npts = int(np.ceil(abs(self.angle/self.dang)))
        R = self.radius+np.sign(self.angle)*self.width/2
        print(R, self.circ_origin, self._xhat, self._yhat)
        for i in range(Npts):
            ang = abs((i+1)/Npts * self.angle * np.pi/180)
            pnew = self.circ_origin + R*(self._xhat*np.cos(ang)+self._yhat*np.sin(ang))
            points.append((pnew[0], pnew[1]))
        
        return points[::-1]

    
############################################################
#                    THE STRIP PATH CLASS                  #
############################################################

class StripPath:

    def __init__(self, pcb: PCB):
        self.pcb: PCB = pcb
        self.path: list[RouteElement] = []
        self.z: float = 0

    def _has(self, element: RouteElement) -> bool:
        if element in self.path:
            return True
        return False
    
    @property
    def xs(self) -> list[float]:
        return [elem.x for elem in self.path]
    
    @property
    def ys(self) -> list[float]:
        return [elem.y for elem in self.path]

    @property
    def start(self) -> RouteElement:
        """ The start of the stripline. """
        return self.path[0]
    
    @property
    def end(self) -> RouteElement:
        """ The end of the stripline """
        return self.path[-1]
    
    def _check_loops(self) -> None:
        if self.path[0]==self.path[-1]:
            raise RouteException('Loops are currently not supported. To fix this problem, implement a single .cut() call before a .straight() call to break the loop.')
        return None
    
    def init(self, 
             x: float, 
             y: float, 
             width: float, 
             direction: tuple[float, float],
             z: float = 0) -> StripPath:
        """ Initializes the StripPath object for routing. """
        self.path.append(StripLine(x, y, width, direction))
        self.z = z
        return self
    
    def _add_element(self, element: RouteElement) -> StripPath:
        """ Adds the provided RouteElement to the path. """
        self.path.append(element)
        self._check_loops()
        return self
    
    def straight(self, distance: 
                 float, width: float | None = None, 
                 dx: float = 0, 
                 dy: float = 0) -> StripPath:
        """Add A straight section to the stripline.

        Adds a straight section with a length determined by "distance". Optionally, a 
        different "width" can be provided. The start of the straight section will be
        at the end of the last section. The optional dx, dy arguments can be used to offset
        the starting coordinate of the straight segment.

        Args:
            distance (float): The length of the stripline
            width (float, optional): The width of the stripline. Defaults to None.
            dx (float, optional): An x-direction offset. Defaults to 0.
            dy (float, optional): A y-direction offset. Defaults to 0.

        Returns:
            StripPath: The current StripPath object.
        """
        
        x = self.end.x + dx
        y = self.end.y + dy

        dx_2, dy_2 = self.end.direction
        x1 = x + distance * dx_2
        y1 = y + distance * dy_2

        if width is not None:
            if width != self.end.width:
                self._add_element(StripLine(x, y, width, (dx_2, dy_2)))

        self._add_element(StripLine(x1, y1, self.end.width, (dx_2, dy_2)))
        return self
    
    def taper(self, distance: float, 
                 width: float) -> StripPath:
        """Add A taper section to the stripline.

        Adds a taper section with a length determined by "distance". Optionally, a 
        different "width" can be provided. The start of the straight section will be
        at the end of the last section. The optional dx, dy arguments can be used to offset
        the starting coordinate of the straight segment.

        Args:
            distance (float): The length of the stripline
            width (float, optional): The width of the stripline. Defaults to None.
            dx (float, optional): An x-direction offset. Defaults to 0.
            dy (float, optional): A y-direction offset. Defaults to 0.

        Returns:
            StripPath: The current StripPath object.
        """
        
        x = self.end.x 
        y = self.end.y 

        dx_2, dy_2 = self.end.direction
        x1 = x + distance * dx_2
        y1 = y + distance * dy_2

        self._add_element(StripLine(x1, y1, width, (dx_2, dy_2)))
        
        return self
    
    def turn(self, angle: float, 
             width: float | None = None, 
             corner_type: Literal['champher','square'] = 'champher') -> StripPath:
        """Adds a turn to the strip path.

        The angle is specified in degrees. The width of the turn will be the same as the last segment.
        optionally, a different width may be provided. 
        By default, all corners will be cut using the "champher" type. Other options are not yet provided.

        Args:
            angle (float): The turning angle
            width (float, optional): The stripline width. Defaults to None.
            corner_type (str, optional): The corner type. Defaults to 'champher'.

        Returns:
            StripPath: The current StripPath object
        """
        x, y = self.end.x, self.end.y
        dx, dy = self.end.direction
        if abs(angle) <= 20:
            corner_type = 'square'
            logger.warning('Small turn detected, defaulting to rectangular corners because chamfers would add to much detail.')
        if width is not None:
            if width != self.end.width:
                self._add_element(StripLine(x, y, width, (dx, dy)))
        else:
            width=self.end.width
        self._add_element(StripTurn(x, y, width, (dx, dy), angle, corner_type))
        return self

    def curve(self, angle: float, radius: float,
             width: float | None = None, 
             corner_type: Literal['champher','square'] = 'champher',
             dang: float = 10) -> StripPath:
        """Adds a bend to the strip path.

        The angle is specified in degrees. The width of the turn will be the same as the last segment.
        optionally, a different width may be provided. 
        By default, all corners will be cut using the "champher" type. Other options are not yet provided.

        Args:
            angle (float): The turning angle
            width (float, optional): The stripline width. Defaults to None.
            corner_type (str, optional): The corner type. Defaults to 'champher'.

        Returns:
            StripPath: The current StripPath object
        """
        if angle == 0:
            logger.trace('Zero angle turn, passing action')
            return self
        x, y = self.end.x, self.end.y
        dx, dy = self.end.direction
        if width is not None:
            if width != self.end.width:
                self._add_element(StripLine(x, y, width, (dx, dy)))
        else:
            width=self.end.width
        self._add_element(StripCurve(x, y, width, (dx, dy), angle, radius, dang=dang))
        return self
    
    def store(self, name: str) -> StripPath:
        """ Store the current x,y coordinate labeled in the PCB object.

        The stored coordinate can be accessed by calling the .load() method on the PCBRouter class.

        Args:
            name (str): The coordinate label

        Returns:
            StripPath: The current StripPath object.
        """
        self.pcb.stored_striplines[name] = self.end
        return self
    
    def split(self, 
              direction: tuple[float, float]  | None= None,
              width: float  | None= None) -> StripPath:
        """Split the current path in N new paths given by a new departure direction

        Args:
            directions (list[tuple[float, float]]): a list of directions example: [(1,0),(-1,0)]
            widths (list[float], optional): The width for each new path. Defaults to None.

        Returns:
            list[StripPath]: A list of new StripPath objects
        """
        if width is None:
            width = self.end.width
        if direction is None:
            direction = self.end.direction
        x = self.end.x
        y = self.end.y
        z = self.z
        paths = self.pcb.new(x,y,width, direction, z)
        self.pcb._checkpoint = self
        return paths

    def lumped_element(self, impedance_function: Callable, size: SIZE_NAMES | tuple) -> StripPath:
        """Adds a lumped element to the PCB.

        The first argument should be the impedance function as function of frequency. For a capacitor this would be:
        Z(f) = 1/(j2πfC).
        The second argument specifies the size of the element (length x width) as a tuple or it can be a string for a
        package. For example "0402". The size of the lumped component does not inlcude the footprint.

        For example a 0602 pacakge has timensions: length=0.6mm, width=0.3mm. The actual length of the component
        not overlapping with the solder pad is 0.3mm (always half) so the component size added is 0.3mm x 0.3mm.
        
        After creation, the trace continues after the lumped component.
        
        You can add the components to your model as following:

        >>> lumped_elements = pcb.lumped_elments
        for le in lumped_elements:
            model.mw.bc.LumpedElement(le)
        
        The impedance function and geometry is automatically passed on with the lumped element added.

        Args:
            impedance_function (Callable): A function that computes the component impedance as a function of frequency.
            size (SizeNames | tuple): The dimensions of the lumped element on PCB.

        Returns:
            StripPath: The same strip path object
        """
        if size in _SMD_SIZE_DICT:
            length, width = _SMD_SIZE_DICT[size]
        else:
            length, width = size
        
        dx, dy = self.end.direction
        x, y = self.end.x, self.end.y
        rx, ry = self.end.dirright
        wh = width/2
        xs = np.array([x+rx*wh, x+rx*wh+length*dx, x-rx*wh+length*dx, x-rx*wh])*self.pcb.unit
        ys = np.array([y+ry*wh, y+ry*wh+length*dy, y-ry*wh+length*dy, y-ry*wh])*self.pcb.unit
        poly = XYPolygon(xs, ys)
        
        self.pcb._lumped_element(poly, impedance_function, width, length)
        return self.pcb.new(x+dx*length, y+dy*length, self.end.width, self.end.direction, self.z)

    def cut(self) -> StripPath:
        """Split the current path in N new paths given by a new departure direction

        Args:
            directions (list[tuple[float, float]]): a list of directions example: [(1,0),(-1,0)]
            widths (list[float], optional): The width for each new path. Defaults to None.

        Returns:
            list[StripPath]: A list of new StripPath objects
        """
        width = self.end.width
        direction = self.end.direction
        x = self.end.x
        y = self.end.y
        z = self.z
        paths = self.pcb.new(x,y,width, direction, z)
        return paths
    
    def stub(self, direction: tuple[float, float],
             width: float,
             length: float, 
             mirror: bool = False) -> StripPath:
        """ Add a single rectangular strip line section at the current coordinate"""
        self.pcb.new(self.end.x, self.end.y, width, direction, self.z).straight(length)
        if mirror:
            self.pcb.new(self.end.x, self.end.y, width, (-direction[0], -direction[1]), self.z).straight(length)
        return self
    
    def merge(self) -> StripPath:
        """Continue at the last point where .split() is called"""
        if self.pcb._checkpoint is None:
            raise RouteException('No checkpoint known. Make sure to call .check() first')
        return self.pcb._checkpoint
    
    def via(self,
            znew: float,
            radius: float,
            proceed: bool = True,
            direction: tuple[float, float] | None = None,
            width: float | None = None,
            extra: float | None = None,
            segments: int = 6) -> StripPath:
        """Adds a via to the circuit

        If proceed is set to True, a new StripPath will be started. The width and direction properties
        will be inherited from the current one if not specified.
        The extra parameter specifies how much extra stripline is added beyond the current point and before
        the new segment to include the via. If not specifies it defaults to width/2.
        
        Args:
            znew (float): The new Z-height for the stripline
            radius (float): The via radius
            proceed (bool, optional): Wether to continue with a new trace. Defaults to True.
            direction (tuple[float, float], optional): The new direction. Defaults to None.
            width (float, optional): The new width. Defaults to None.
            extra (float, optional): How much extra stripline to add around the via. Defaults to None.
            segments (int, optional): The number of via polygon sections. Defaults to 6.

        Returns:
            StripPath: The new StripPath object
        """
        
        if extra is None:
            extra = self.end.width/2
        x, y = self.end.x, self.end.y
        z1 = self.z
        z2 = znew
        if extra > 0:
            self.straight(extra)
        self.pcb.vias.append(Via(x,y,z1,z2,radius, segments))
        if proceed:
            if width is None:
                width = self.end.width
            if direction is None:
                direction = self.end.direction
            dx = direction[0]*extra
            dy = direction[1]*extra
            return self.pcb.new(x-dx, y-dy, width, direction, z2)
        return self
    
    def short(self) -> StripPath:
        self.via(self.pcb.z(1), self.end.width/3, False)
        return self

    def jump(self, 
             dx: float | None = None,
             dy: float | None = None,
             width: float | None = None,
             direction: tuple[float, float] | None = None,
             gap: float | None = None,
             side: Literal['left','right'] | None = None,
             reverse: float | None = None) -> StripPath:
        """Add an unconnected jump to the currenet stripline.

        The last stripline path will be terminated and a new one will be started based on the 
        displacement provided by dx and dy. The new path will proceed in the same direction or
        another one based ont the "direction" argument.
        An alternative one can define a "gap", "side" and "reverse" argument. The stripline
        will make a lateral jump ensuring a gap between the current and new line. The direction
        of the jump is either "left" or "right" as seen from the direction of the stripline.
        The reverse argument is a distance by which the stripline moves back.

        Args:
            dx (float, optional): The jumps dx distance. Defaults to None.
            dy (float, optional): The jumps dy distance. Defaults to None.
            width (float, optional): The new stripline width. Defaults to None.
            direction (tuple[float, float], optional): The new stripline direction. Defaults to None.
            gap (float, optional): The gap between the current and next stripline. Defaults to None.
            side (Literal[left, right], optional): The lateral jump direction. Defaults to None.
            reverse (float, optional): How much to move back if a lateral jump is made. Defaults to None.

        Example:
        The current example would yield a coupled line filter parallel jump.
        >>> StripPath.jump(gap=1, side="left", reverse=quarter_wavelength).straight(...)

        Returns:
            StripPath: The new StripPath object
        """
        if width is None:
            width = self.end.width
        if direction is None:
            direction = self.end.direction
        else:
            direction = np.array(direction)

        ending = self.end

        if gap is not None and side is not None and reverse is not None:
            Q = 1
            if side=='left':
                Q = -1
            x = ending.x - reverse*ending.direction[0] + Q*ending.dirright[0]*(width/2 + ending.width/2 + gap)
            y = ending.y - reverse*ending.direction[1] + Q*ending.dirright[1]*(width/2 + ending.width/2 + gap)
        else:
            x = ending.x + dx
            y = ending.y + dy
        return self.pcb.new(x, y, width, direction)
    
    def to(self, dest: tuple[float, float], 
           arrival_dir: tuple[float, float] | None = None,
           arrival_margin: float  | None= None,
           angle_step: float = 90) -> StripPath:
        """
        Extend the path from current end point to dest (x, y).
        Optionally ensure arrival in arrival_dir after a straight segment of arrival_margin.
        Turns are quantized to multiples of angle_step (divisor of 360, <=90).
        """
        # Validate angle_step
        if 360 % angle_step != 0 or angle_step > 90 or angle_step <= 0:
            raise ValueError(f"angle_step must be a positive divisor of 360 <= 90, got {angle_step}")

        # Current state
        x0, y0 = self.end.x, self.end.y
        vx, vy = self.end.direction  # unit heading
        tx, ty = dest

        # Compute unit arrival direction
        if arrival_dir is not None:
            adx, ady = arrival_dir
            mag = math.hypot(adx, ady)
            if mag == 0:
                raise ValueError("arrival_dir must be non-zero")
            ux, uy = adx/mag, ady/mag
        else:
            # if no arrival_dir, just head to point
            ux, uy = 0.0, 0.0
        arrival_margin += self.end.width
        # Compute base point: destination minus arrival margin
        bx = tx - ux * arrival_margin
        by = ty - uy * arrival_margin

        # Parametric search along negative arrival direction
        # we seek t >= 0 such that angle from (vx,vy) to (bx - x0 - ux*t, by - y0 - uy*t)
        # quantizes exactly to a multiple of angle_step
        atol = angle_step/10
        dtol = 0.01
        t = 0.0
        max_t = math.hypot(bx - x0, by - y0) + arrival_margin + 1e-3
        dt = max_t / 1000.0  # resolution of search
        found = False
        desired_q = None
        #cand_dx = cand_dy = 0.0
        
        while t <= max_t:
            # candidate intercept point
            cx = bx - ux * t
            cy = by - uy * t
            dx = cx - x0
            dy = cy - y0
            if abs(dx) < dtol and abs(dy) < dtol:
                # reached start; skip
                t += dt
                continue
            # compute angle
            cross = vx*dy - vy*dx
            dot = vx*dx + vy*dy
            ang = math.degrees(math.atan2(cross, dot))
            # quantize
            q_ang = math.ceil(ang / angle_step) * angle_step
            if abs(ang - q_ang) <= atol:
                found = True
                desired_q = q_ang
                break
            t += dt

        if not found:
            raise RuntimeError("Could not find an intercept angle matching quantization")

        # 1) Perform initial quantized turn
        if abs(desired_q) > atol: # type: ignore
            self.turn(-desired_q) # type: ignore
        x0 = self.end.x
        y0 = self.end.y
        # compute new heading vector after turn
        theta = math.radians(desired_q)
        nvx = math.cos(theta) * vx - math.sin(theta) * vy
        nvy = math.sin(theta) * vx + math.cos(theta) * vy

        # 2) Compute exact intercept distance via line intersection:
        # Solve: (x0,y0) + s*(nvx,nvy) = (bx,by) - t*(ux,uy)
        # Unknowns s,t (we reuse t from above as initial guess, but solve fresh):
        tol_dist = 1e-6
        A11, A12 = nvx, ux
        A21, A22 = nvy, uy
        B1 = bx - x0
        B2 = by - y0
        det = A11 * A22 - A12 * A21
        if abs(det) < tol_dist:
            raise RuntimeError("Initial heading parallel to arrival line, no unique intercept")
        s = (B1 * A22 - B2 * A12) / det
        t_exact = (A11 * B2 - A21 * B1) / det
        if s < -tol_dist or (arrival_dir is not None and t_exact < -tol_dist):
            raise RuntimeError("Computed intercept lies behind start or before arrival point")

        
        # 3) Turn into arrival direction (if provided)
    
        # we need to rotate from current heading (vx,vy) by desired_q to (nvx,nvy)
        theta = math.radians(desired_q)
        nvx = math.cos(theta)*vx - math.sin(theta)*vy
        nvy = math.sin(theta)*vx + math.cos(theta)*vy
        # target heading is (ux,uy)
        cross2 = nvx*uy - nvy*ux
        dot2 = nvx*ux + nvy*uy
        back_ang = math.degrees(math.atan2(cross2, dot2))
        
        backoff = math.tan(abs(back_ang)*np.pi/360)*self.end.width/2

        self.straight(s - backoff)
        self.turn(-back_ang)

        x0 = self.end.x
        y0 = self.end.y
        D = math.hypot(tx-x0, ty-y0)
        # 4) Final straight into destination by arrival_margin + t
        self.straight(D)

        return self

    def macro(self, path: str, width: float | None = None, start_dir: tuple[float, float] | None = None) -> StripPath:
        r"""Parse an EMerge macro command string

        The start direction by default is the abslute current heading. If a specified heading is provided
        the macro language will assume that as the current heading and generate commands accordingly. 

        The language is specified by a symbol plus a number.
        Symbols:
        - = X: Move X units forward
        - \> X: Turn to right and move X forward
        - < X: Turn to left and move X forward
        - v X: Turn to down and move X forward
        - ^ X: Turn to up and move X forward
        - T X,Y: Taper X forward to width Y
        - \\ X: Turn relative right 90 degrees and X forward
        - / X: Turn relative left 90 degrees and X forward

        (*) All commands X can also be provided as X,Y to change the width

        Args:
            path (str): The path command string
            width (float, optional): The width to start width. Defaults to None.
            start_dir (tuple[float, float], optional): The start direction to assume. Defaults to None.

        Example:
        >>> my_pcb.macro("= 5 v 4,1.2 > 5 ^ 2 > 3 T 4, 2.1")
        
        Returns:
            StripPath: The strippath object
        """
        if start_dir is None:
            start_dir = self.end.direction
        if width is None:
            width = self.end.width
        for instr in parse_macro(path, width, start_dir):
            getattr(self, instr.instr)(*instr.args, **instr.kwargs)
        return self
    
    def __call__(self, element_nr: int) -> RouteElement:
        if element_nr >= len(self.path):
            self.path.append(RouteElement())
        return self.path[element_nr]

############################################################
#                     PCB DESIGN CLASS                     #
############################################################

class PCB:
    def __init__(self,
                 thickness: float,
                 unit: float = 0.001,
                 cs: CoordinateSystem | None = None,
                 material: Material = AIR,
                 layers: int = 2,
                 ):

        self.thickness: float = thickness
        self._zs: np.ndarray = np.linspace(-self.thickness, 0, layers)
        self.material: Material = material
        self.width: float | None = None
        self.length: float | None = None
        self.origin: np.ndarray = np.array([0.,0.,0.])
        self.paths: list[StripPath] = []
        self.polies: list[PCBPoly] = []

        self.lumped_ports: list[StripLine] = []
        self.lumped_elements: list[GeoPolygon] = []

        self.unit: float = unit

        self.cs: CoordinateSystem = cs
        if self.cs is None:
            self.cs = GCS

        self.traces: list[GeoPolygon] = []
        self.ports: list[GeoPolygon] = []
        self.vias: list[Via] = []

        self.xs: list[float] = []
        self.ys: list[float] = []
        self.zs: list[float] = []

        self.stored_coords: dict[str,tuple[float, float]] = dict()
        self.stored_striplines: dict[str, StripLine] = dict()
        self._checkpoint: StripPath = None

        self.calc: PCBCalculator = PCBCalculator(self.thickness, self._zs, self.material, self.unit)

    @property
    def trace(self) -> GeoPolygon:
        tags = []
        for trace in self.traces:
            tags.extend(trace.tags)
        return GeoPolygon(tags)
    
    
    @property
    def all_objects(self) -> list[GeoPolygon]:
        return self.traces + self.ports
    

    def z(self, layer: int) -> float:
        """Returns the z-height of the given layer number counter from 1 (bottom) to N (top)

        Args:
            layer (int): The layer number (1 to N)

        Returns:
            float: the z-height
        """
        return self._zs[layer-1]
    
    def _get_z(self, element: RouteElement) -> float :
        """Return the z-height of a given Route Element

        Args:
            element (RouteElement): The requested route element

        Returns:
            float: The z-height.
        """
        for path in self.paths:
            if path._has(element):
                return path.z
        raise RouteException('Requesting z-height of route element that is not contained in a path.')

    def load(self, name: str) -> StripLine:
        """Acquire the x,y, coordinate associated with the label name.
        
        Args:
            name (str): The name of the x,y coordinate
            
        """
        if name in self.stored_striplines:
            return self.stored_striplines[name]
        else:
            raise ValueError(f'There is no stripline or coordinate under the name of {name}')
    
    def __call__(self, path_nr: int) -> StripPath:
        if path_nr >= len(self.paths):
            self.paths.append(StripPath(self))
        return self.paths[path_nr]
    
    def determine_bounds(self, 
                         leftmargin: float = 0,
                         topmargin: float = 0,
                         rightmargin: float = 0,
                         bottommargin: float = 0):
        """Defines the rectangular boundary of the PCB.

        Args:
            leftmargin (float, optional): The left margin. Defaults to 0.
            topmargin (float, optional): The top margin. Defaults to 0.
            rightmargin (float, optional): The right margin. Defaults to 0.
            bottommargin (float, optional): The bottom margin. Defaults to 0.
        """
        if len(self.xs) == 0:
            raise ValueError('PCB path is not compiled. Compile before defining boundaries.')
        minx = min(self.xs)
        maxx = max(self.xs)
        miny = min(self.ys)
        maxy = max(self.ys)
        ml = leftmargin
        mt = topmargin
        mr = rightmargin
        mb = bottommargin
        self.width = (maxx - minx + mr + ml)
        self.length = (maxy - miny + mt + mb)
        self.origin = np.array([-ml+minx, -mb+miny, 0])
    
    def set_bounds(self,
                   xmin: float,
                   ymin: float,
                   xmax: float,
                   ymax: float) -> None:
        """Define the bounds of the PCB

        Args:
            xmin (float): The minimum x-coordinate
            ymin (float): The minimum y-coordinate
            xmax (float): The maximum x-coordinate
            ymax (float): The maximum y-coordinate
        """
        self.origin = np.array([xmin, ymin, 0])
        self.width = xmax-xmin
        self.length = ymax-ymin

    def plane(self,
              z: float,
              width: float | None = None,
              height: float | None = None,
              origin: tuple[float, float] | None = None,
              alignment: Literal['corner','center'] = 'corner') -> GeoSurface:
        """Generates a generic rectangular plate in the XY grid.
        If no size is provided, it defaults to the entire PCB size assuming that the bounds are determined.

        Args:
            z (float): The Z-height for the plate.
            width (float, optional): The width of the plate. Defaults to None.
            height (float, optional): The height of the plate. Defaults to None.
            origin (tuple[float, float], optional): The origin of the plate. Defaults to None.
            alignment (['corner','center], optional): The alignment of the plate. Defaults to 'corner'.

        Returns:
            GeoSurface: _description_
        """
        if width is None or height is None or origin is None:
            if self.width is None or self.length is None or self.origin:
                raise RouteException('Cannot define a plane with no possible definition of its size.')
            width = self.width
            height = self.length
            origin = (self.origin[0]*self.unit, self.origin[1]*self.unit)
        
        origin: tuple[float, ...] = origin + (z*self.unit, ) # type: ignore

        if alignment == 'center':
            origin = (origin[0] - width*self.unit/2, 
                                                        origin[1] - height*self.unit/2, 
                                                        origin[2])

        plane = Plate(origin, (width*self.unit, 0, 0), (0, height*self.unit, 0)) # type: ignore
        plane = change_coordinate_system(plane, self.cs) # type: ignore
        return plane # type: ignore
    
    def gen_pcb(self, 
                split_z: bool = True,
                layer_tolerance: float = 1e-6,
                merge: bool = True) -> GeoVolume:
        """Generate the PCB Block object

        Returns:
            GeoVolume: The PCB Block
        """
        x0, y0, z0 = self.origin*self.unit

        if split_z:
            zvalues = sorted(list(set(self.zs + [-self.thickness, 0.0])))
            zvalues_isolated = [zvalues[0],]
            for z in zvalues[1:]:
                if (z-zvalues_isolated[-1]) <= layer_tolerance:
                    continue
                zvalues_isolated.append(z)
            boxes: list[GeoVolume] = []
            for z1, z2 in zip(zvalues_isolated[:-1],zvalues_isolated[1:]):
                h = z2-z1
                box = Box(self.width*self.unit,
                          self.length*self.unit,
                          h*self.unit,
                          position=(x0, y0, z0+z1*self.unit))
                box.material = self.material
                box = change_coordinate_system(box, self.cs)
                boxes.append(box)
            if merge:
                return GeoVolume.merged(boxes) # type: ignore
            return boxes # type: ignore
        
        box = Box(self.width*self.unit, 
                  self.length*self.unit, 
                  self.thickness*self.unit, 
                  position=(x0,y0,z0-self.thickness*self.unit))
        box.material = self.material
        box = change_coordinate_system(box, self.cs)
        return box # type: ignore

    def gen_air(self, height: float) -> GeoVolume:
        """Generate the Air Block object

        This requires that the width, depth and origin are deterimed. This 
        can either be done manually or via the .determine_boudns() method.

        Returns:
            GeoVolume: The PCB Block
        """
        x0, y0, z0 = self.origin*self.unit
        box = Box(self.width*self.unit, 
                  self.length*self.unit, 
                  height*self.unit, 
                  position=(x0,y0,z0))
        box = change_coordinate_system(box, self.cs)
        return box # type: ignore

    def new(self, 
            x: float, 
            y: float, 
            width: float, 
            direction: tuple[float, float],
            z: float = 0) -> StripPath:
        """Start a new trace

        The trace is started at the provided x,y, coordinates with a width "width".
        The direction must be provided as an (dx,dy) vector provided as tuple.

        Args:
            x (float): The starting X-coordinate (local)
            y (float): The starting Y-coordinate (local)
            width (float): The (micro)-stripline width
            direction (tuple[float, float]): The direction.

        Returns:
            StripPath: A StripPath object that can be extended with method chaining.

        Example:
        >>> PCB.new(...).straight(...).turn(...).straight(...) etc.

        """
        path = StripPath(self)
        path.init(x, y, width, direction, z=z)
        self.paths.append(path)  
        return path
    
    def lumped_port(self, stripline: StripLine, z_ground: float | None = None) -> GeoPolygon:
        """Generate a lumped-port object to be created.

        Args:
            stripline (StripLine): _description_
        """
        
        xy1 = stripline.right[0]
        xy2 = stripline.left[0]
        z = self._get_z(stripline)
        if z_ground is None:
            z_ground = -self.thickness
        height = z-z_ground
        x1, y1, z1 = self.cs.in_global_cs(xy1[0]*self.unit, xy1[1]*self.unit, z*self.unit - height*self.unit)
        x2, y2, z2 = self.cs.in_global_cs(xy1[0]*self.unit, xy1[1]*self.unit, z*self.unit )
        x3, y3, z3 = self.cs.in_global_cs(xy2[0]*self.unit, xy2[1]*self.unit, z*self.unit )
        x4, y4, z4 = self.cs.in_global_cs(xy2[0]*self.unit, xy2[1]*self.unit, z*self.unit - height*self.unit)
        
        ptag1 = gmsh.model.occ.addPoint(x1, y1, z1)
        ptag2 = gmsh.model.occ.addPoint(x2, y2, z2)
        ptag3 = gmsh.model.occ.addPoint(x3, y3, z3)
        ptag4 = gmsh.model.occ.addPoint(x4, y4, z4)
        
        ltag1 = gmsh.model.occ.addLine(ptag1, ptag2)
        ltag2 = gmsh.model.occ.addLine(ptag2, ptag3)
        ltag3 = gmsh.model.occ.addLine(ptag3, ptag4)
        ltag4 = gmsh.model.occ.addLine(ptag4, ptag1)
        
        ltags = [ltag1, ltag2, ltag3, ltag4]
        
        tag_wire = gmsh.model.occ.addWire(ltags)
        planetag = gmsh.model.occ.addPlaneSurface([tag_wire,])
        poly = GeoPolygon([planetag,])
        poly._aux_data['width'] = stripline.width*self.unit
        poly._aux_data['height'] = height*self.unit
        poly._aux_data['vdir'] = self.cs.zax
        poly._aux_data['idir'] = Axis(self.cs.xax.np*stripline.dirright[0] + self.cs.yax.np*stripline.dirright[1])
        
        return poly

    def _lumped_element(self, poly: XYPolygon, function: Callable, width: float, length: float) -> None:

        geopoly = poly._finalize(self.cs)
        geopoly._aux_data['func'] = function
        geopoly._aux_data['width'] = width
        geopoly._aux_data['height'] = length
        self.lumped_elements.append(geopoly)


    def modal_port(self,
                  point: StripLine,
                  height: float,
                  width_multiplier: float = 5.0,
                  ) -> GeoSurface:
        """Generate a wave-port as a GeoSurface.

        The port is placed at the coordinate of the provided stripline. The width
        is determined as a multiple of the stripline width. The height will be 
        extended to the air height from the bottom of the PCB unless a different height is specified.

        Args:
            point (StripLine): The location of the port.
            width_multiplier (float, optional): The width of the port in stripline widths. Defaults to 5.0.
            height (float, optional): The height of the port. Defaults to None.

        Returns:
            GeoSurface: The GeoSurface object that can be used for the waveguide.
        """
        
        height = (self.thickness + height)
        
        ds = point.dirright
        x0 = point.x - ds[0]*point.width*width_multiplier/2
        y0 = point.y - ds[1]*point.width*width_multiplier/2
        z0 =  - self.thickness
        ax1 = np.array([ds[0], ds[1], 0])*self.unit*point.width*width_multiplier
        ax2 = np.array([0,0,1])*height*self.unit

        plate = Plate(np.array([x0,y0,z0])*self.unit, ax1, ax2)
        plate = change_coordinate_system(plate, self.cs)
        return plate # type: ignore

    def generate_vias(self, merge=False) -> list[Cyllinder] | Cyllinder:
        """Generates the via objects.

        Args:
            merge (bool, optional): Whether to merge the result into a final object. Defaults to False.

        Returns:
            list[Cyllinder] | Cyllinder: Either al ist of cylllinders or a single one (merge=True)
        """
        vias = []
        for via in self.vias:
            x0 = via.x*self.unit
            y0 = via.y*self.unit
            z0 = via.z1*self.unit
            xg, yg, zg = self.cs.in_global_cs(x0, y0, z0)
            cs = CoordinateSystem(self.cs.xax, self.cs.yax, self.cs.zax, np.array([xg, yg, zg]))
            cyl = Cyllinder(via.radius*self.unit, (via.z2-via.z1)*self.unit, cs, via.segments)
            cyl.material = COPPER
            vias.append(cyl)
        if merge:
            
            return GeoVolume.merged(vias) # type: ignore
        return vias   

    def add_poly(self, 
                 xs: list[float],
                 ys: list[float],
                 z: float = 0,
                 material: Material = COPPER) -> None:
        """Add a custom polygon to the PCB

        Args:
            xs (list[float]): A list of x-coordinates
            ys (list[float]): A list of y-coordinates
            z (float, optional): The z-height. Defaults to 0.
            material (Material, optional): The material. Defaults to COPPER.
        """
        self.polies.append(PCBPoly(xs, ys, z, material))

    def _gen_poly(self, xys: list[tuple[float, float]], z: float) -> GeoPolygon:
        """ Generates a GeoPoly out of a list of (x,y) coordinate tuples"""
        ptags = []
        for x,y in xys:
            px, py, pz = self.cs.in_global_cs(x*self.unit, y*self.unit, z*self.unit)
            ptags.append(gmsh.model.occ.addPoint(px, py, pz))
        
        ltags = []
        for t1, t2 in zip(ptags[:-1], ptags[1:]):
            ltags.append(gmsh.model.occ.addLine(t1, t2))
        ltags.append(gmsh.model.occ.addLine(ptags[-1], ptags[0]))
        
        tag_wire = gmsh.model.occ.addWire(ltags)
        planetag = gmsh.model.occ.addPlaneSurface([tag_wire,])
        poly = GeoPolygon([planetag,])
        return poly
    
    @overload
    def compile_paths(self, merge: Literal[True]) -> GeoSurface: ...
    
    @overload
    def compile_paths(self, merge: Literal[False] = ...) -> list[GeoSurface]: ...

    def compile_paths(self, merge: bool = False) -> list[GeoPolygon] | GeoSurface:
        """Compiles the striplines and returns a list of polygons or asingle one.

        The Z=0 argument determines the height of the striplines. Z=0 corresponds to the top of
        the PCB.

        Args:
            merge (bool, optional): Whether to merge the Polygons into a single. Defaults to False.

        Returns:
            list[Polygon] | GeoSurface: The output stripline polygons possibly merged if merge = True.
        """
        polys = []
        allx = []
        ally = []

        for path in self.paths:
            z = path.z
            self.zs.append(z)
            xys = []
            for elemn in path.path:
                xys.extend(elemn.right)
            for element in path.path[::-1]:
                xys.extend(element.left)

            xm, ym = xys[0]
            xys2 = [(xm,ym),]
            
            for x,y in xys[1:]:
                if ((x-xm)**2 + (y-ym)**2)>1e-6:
                    xys2.append((x,y))
                    xm, ym = x, y
                    allx.append(x)
                    ally.append(y)
            
            poly = self._gen_poly(xys2, z)
            poly.material = COPPER
            polys.append(poly)

        for pcbpoly in self.polies:
            self.zs.append(pcbpoly.z)
            poly = self._gen_poly(pcbpoly.xys, pcbpoly.z)
            poly.material = pcbpoly.material
            polys.append(poly)

        self.xs = allx
        self.ys = ally

        self.traces = polys
        if merge:
            tags = []
            for p in polys:
                tags.extend(p.tags)
                if p.material != COPPER:
                    logger.warning(f'Merging a polygon with material {p.material} into a single polygon that will be COPPER.')
            polys = GeoSurface(tags)
            polys.material = COPPER
        return polys
                
############################################################
#                        DEPRICATED                       #
############################################################

class PCBLayouter(PCB):

    def __init__(self, *args, **kwargs):
        logger.warning('PCBLayouter will be depricated. Use PCB instead.')
        super().__init__(*args, **kwargs)
        


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

from .microwave_bc import PortBC
from ...mth.integrals import surface_integral
import numpy as np
from typing import Callable

def sparam_waveport(nodes: np.ndarray,
                    tri_vertices: np.ndarray,
                    bc: PortBC, 
                    freq: float,
                    fieldf: Callable,
                    ndpts: int = 4):
    ''' Compute the S-parameters assuming a wave port mode
    
    Arguments:
    ----------
    nodes: np.ndarray = (3,:) np.ndarray of all nodes in the mesh.
    tri_vertices: np.ndarray = (3,:) np.ndarray of triangle indices that need to be integrated,
    bc: RobinBC = The port boundary condition object
    freq: float = The frequency at which to do the calculation
    fielf: Callable = The interpolation fuction that computes the E-field from the simulation
    ndpts: int = 4 the number of Duvanant integration points to use (default = 4)
    '''
    
    def modef(x, y, z):
        return bc.port_mode_3d_global(x, y, z, freq)

    def modef_c(x, y, z):
        return np.conj(modef(x, y, z))
    
    Q = 0
    if bc.active:
        Q = 1

    def fieldf_p(x, y, z):
        return fieldf(x,y,z) - Q * modef(x,y,z)
    
    
    def inproduct1(x, y, z):
        Ex1, Ey1, Ez1 = fieldf_p(x,y,z)
        Ex2, Ey2, Ez2 = modef_c(x,y,z)
        return Ex1*Ex2 + Ey1*Ey2 + Ez1*Ez2
    
    def inproduct2(x, y, z):
        Ex1, Ey1, Ez1 = modef(x,y,z)
        Ex2, Ey2, Ez2 = modef_c(x,y,z)
        return Ex1*Ex2 + Ey1*Ey2 + Ez1*Ez2
    
    mode_dot_field = surface_integral(nodes, tri_vertices, inproduct1, gq_order=ndpts)
    norm = surface_integral(nodes, tri_vertices, inproduct2, gq_order=ndpts)
    
    svec = mode_dot_field/norm
    return svec

def sparam_mode_power(nodes: np.ndarray,
                    tri_vertices: np.ndarray,
                    bc: PortBC, 
                    k0: float,
                    const: np.ndarray,
                    gq_order: int = 4):
    ''' Compute the S-parameters assuming a wave port mode
    
    Arguments:
    ----------
    nodes: np.ndarray = (3,:) np.ndarray of all nodes in the mesh.
    tri_vertices: np.ndarray = (3,:) np.ndarray of triangle indices that need to be integrated,
    bc: RobinBC = The port boundary condition object
    freq: float = The frequency at which to do the calculation
    fielf: Callable = The interpolation fuction that computes the E-field from the simulation
    gq_order: int = 4 the Gauss-Quadrature order (default = 4)
    '''

    def modef(x, y, z):
        return bc.port_mode_3d_global(x, y, z, k0)
    
    def inproduct2(x, y, z):
        Ex1, Ey1, Ez1 = modef(x,y,z)
        Ex2, Ey2, Ez2 = np.conj(modef(x,y,z))
        return (Ex1*Ex2 + Ey1*Ey2 + Ez1*Ez2)/(2*bc.Zmode(k0))
    
    norm = surface_integral(nodes, tri_vertices, inproduct2, const, gq_order=gq_order)
    
    return norm

def sparam_field_power(nodes: np.ndarray,
                    tri_vertices: np.ndarray,
                    bc: PortBC, 
                    k0: float,
                    fieldf: Callable,
                    const: np.ndarray,
                    gq_order: int = 4) -> complex:
    ''' Compute the S-parameters assuming a wave port mode
    
    Arguments:
    ----------
    nodes: np.ndarray = (3,:) np.ndarray of all nodes in the mesh.
    tri_vertices: np.ndarray = (3,:) np.ndarray of triangle indices that need to be integrated,
    bc: RobinBC = The port boundary condition object
    freq: float = The frequency at which to do the calculation
    fielf: Callable = The interpolation fuction that computes the E-field from the simulation
    gq_order: int = 4 the Gauss-Quadrature order (default = 4)
    '''
    
    def modef(x, y, z):
        return bc.port_mode_3d_global(x, y, z, k0)
    
    Q = 0
    if bc.active:
        Q = 1

    def fieldf_p(x, y, z):
        return fieldf(x,y,z) - Q * modef(x,y,z)
    
    def inproduct1(x, y, z):
        Ex1, Ey1, Ez1 = fieldf_p(x,y,z)
        Ex2, Ey2, Ez2 = np.conj(modef(x,y,z))
        return (Ex1*Ex2 + Ey1*Ey2 + Ez1*Ez2) / (2*bc.Zmode(k0))
    
    mode_dot_field = surface_integral(nodes, tri_vertices, inproduct1, const, gq_order=gq_order)
    
    svec = mode_dot_field
    return svec


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

import numpy as np
from ....elements import Nedelec2
from scipy.sparse import csr_matrix, coo_matrix
from numba_progress import ProgressBar, ProgressBarType
from ....mth.optimized import local_mapping, matinv, dot_c, cross_c, compute_distances
from numba import c16, types, f8, i8, njit, prange

############################################################
#                  CACHED FACTORIAL VALUES                 #
############################################################

_FACTORIALS = np.array([1, 1, 2, 6, 24, 120, 720, 5040, 40320, 362880], dtype=np.int64)

############################################################
#                  INDEX MAPPING FUNCTIONS                 #
############################################################

# These mapping functions return edge and face coordinates in the appropriate order.

@njit(i8[:,:](i8[:,:], i8[:,:], i8[:,:], i8, i8), cache=True, nogil=True)
def local_tet_to_triid(tet_to_field, tets, tris, itet, nedges) -> np.ndarray:
    """Returns the triangle node indices in the right order given a tet-index"""
    tri_ids = tet_to_field[6:10, itet] - nedges
    global_tri_map = tris[:, tri_ids]
    return local_mapping(tets[:, itet], global_tri_map)

@njit(i8[:,:](i8[:,:], i8[:,:], i8[:,:], i8), cache=True, nogil=True)
def local_tet_to_edgeid(tets, edges, tet_to_field, itet) -> np.ndarray:
    """Returns the edge node indices in the right order given a tet-index"""
    global_edge_map = edges[:, tet_to_field[:6,itet]]
    return local_mapping(tets[:, itet], global_edge_map)

@njit(i8[:,:](i8[:,:], i8[:,:], i8[:,:], i8), cache=True, nogil=True)
def local_tri_to_edgeid(tris, edges, tri_to_field, itri: int) -> np.ndarray:
    """Returns the edge node indices in the right order given a triangle-index"""
    global_edge_map = edges[:, tri_to_field[:3,itri]]
    return local_mapping(tris[:, itri], global_edge_map)

@njit(c16[:](c16[:,:], c16[:]), cache=True, nogil=True)
def matmul(Mat, Vec):
    ## Matrix multiplication of a 3D vector
    Vout = np.empty((3,), dtype=np.complex128)
    Vout[0] = Mat[0,0]*Vec[0] + Mat[0,1]*Vec[1] + Mat[0,2]*Vec[2]
    Vout[1] = Mat[1,0]*Vec[0] + Mat[1,1]*Vec[1] + Mat[1,2]*Vec[2]
    Vout[2] = Mat[2,0]*Vec[0] + Mat[2,1]*Vec[1] + Mat[2,2]*Vec[2]
    return Vout

@njit(f8(i8, i8, i8, i8), cache=True, fastmath=True, nogil=True)
def volume_coeff(a: int, b: int, c: int, d: int):
    """ Computes the appropriate matrix coefficients given a list of
    barycentric coordinate functions mentioned.
    Example:
      - L1^2 * L2 - volume_coeff(1,1,2,0) """
    klmn = np.array([0,0,0,0,0,0,0])
    klmn[a] += 1
    klmn[b] += 1
    klmn[c] += 1
    klmn[d] += 1
    output = (_FACTORIALS[klmn[1]]*_FACTORIALS[klmn[2]]*_FACTORIALS[klmn[3]]
                  *_FACTORIALS[klmn[4]]*_FACTORIALS[klmn[5]]*_FACTORIALS[klmn[6]])/_FACTORIALS[(np.sum(klmn[1:])+3)]
    return output


############################################################
#        PRECOMPUTATION OF INTEGRATION COEFFICIENTS       #
############################################################

NFILL = 5
VOLUME_COEFF_CACHE_BASE = np.zeros((NFILL,NFILL,NFILL,NFILL), dtype=np.float64)
for I in range(NFILL):
    for J in range(NFILL):
        for K in range(NFILL):
            for L in range(NFILL):
                VOLUME_COEFF_CACHE_BASE[I,J,K,L] = volume_coeff(I,J,K,L)

VOLUME_COEFF_CACHE = VOLUME_COEFF_CACHE_BASE


############################################################
#  COMPUTATION OF THE BARYCENTRIC COORDINATE COEFFICIENTS #
############################################################

@njit(types.Tuple((f8[:], f8[:], f8[:], f8))(f8[:], f8[:], f8[:]), cache = True, nogil=True)
def tet_coefficients_bcd(xs: np.ndarray, ys: np.ndarray, zs: np.ndarray) -> tuple[np.ndarray, np.ndarray, np.ndarray, float]:
    """Computes the a,b,c and d coefficients of a tet barycentric coordinate functions and the volume

    Args:
        xs (np.ndarray): The tetrahedron X-coordinates
        ys (np.ndarray): The tetrahedron Y-coordinates
        zs (np.ndrray): The tetrahedron Z-coordinates

    Returns:
        tuple[np.ndarray, np.ndarray, np.ndarray, float]: The a, b, c, d coefficients and volume
    """
    x1, x2, x3, x4 = xs
    y1, y2, y3, y4 = ys
    z1, z2, z3, z4 = zs

    bbs = np.empty((4,), dtype=np.float64)
    ccs = np.empty((4,), dtype=np.float64)
    dds = np.empty((4,), dtype=np.float64)

    V = np.abs(-x1*y2*z3/6 + x1*y2*z4/6 + x1*y3*z2/6 - x1*y3*z4/6 - x1*y4*z2/6 + \
                x1*y4*z3/6 + x2*y1*z3/6 - x2*y1*z4/6 - x2*y3*z1/6 + x2*y3*z4/6 + \
                x2*y4*z1/6 - x2*y4*z3/6 - x3*y1*z2/6 + x3*y1*z4/6 + x3*y2*z1/6 - \
                x3*y2*z4/6 - x3*y4*z1/6 + x3*y4*z2/6 + x4*y1*z2/6 - x4*y1*z3/6 - \
                x4*y2*z1/6 + x4*y2*z3/6 + x4*y3*z1/6 - x4*y3*z2/6)
    
    bbs[0] = -y2*z3 + y2*z4 + y3*z2 - y3*z4 - y4*z2 + y4*z3
    bbs[1] = y1*z3 - y1*z4 - y3*z1 + y3*z4 + y4*z1 - y4*z3
    bbs[2] = -y1*z2 + y1*z4 + y2*z1 - y2*z4 - y4*z1 + y4*z2
    bbs[3] = y1*z2 - y1*z3 - y2*z1 + y2*z3 + y3*z1 - y3*z2
    ccs[0] = x2*z3 - x2*z4 - x3*z2 + x3*z4 + x4*z2 - x4*z3
    ccs[1] = -x1*z3 + x1*z4 + x3*z1 - x3*z4 - x4*z1 + x4*z3
    ccs[2] = x1*z2 - x1*z4 - x2*z1 + x2*z4 + x4*z1 - x4*z2
    ccs[3] = -x1*z2 + x1*z3 + x2*z1 - x2*z3 - x3*z1 + x3*z2
    dds[0] = -x2*y3 + x2*y4 + x3*y2 - x3*y4 - x4*y2 + x4*y3
    dds[1] = x1*y3 - x1*y4 - x3*y1 + x3*y4 + x4*y1 - x4*y3
    dds[2] = -x1*y2 + x1*y4 + x2*y1 - x2*y4 - x4*y1 + x4*y2
    dds[3] = x1*y2 - x1*y3 - x2*y1 + x2*y3 + x3*y1 - x3*y2

    return bbs, ccs, dds, V


############################################################
#              MAIN CURL-CURL MATRIX ASSEMBLY             #
############################################################

def tet_mass_stiffness_matrices(field: Nedelec2,
                           er: np.ndarray, 
                           ur: np.ndarray) -> tuple[csr_matrix, csr_matrix]:
    """Computes the curl-curl Nedelec-2 mass and stiffness matrices

    Args:
        field (Nedelec2): The Nedelec2 Field object
        er (np.ndarray): a 3x3xN array with permittivity tensors
        ur (np.ndarray): a 3x3xN array with permeability tensors

    Returns:
        tuple[csr_matrix, csr_matrix]: The stiffness and mass matrix.
    """
    tets = field.mesh.tets
    tris = field.mesh.tris
    edges = field.mesh.edges
    nodes = field.mesh.nodes

    nT = tets.shape[1]
    tet_to_field = field.tet_to_field
    tet_to_edge = field.mesh.tet_to_edge
    nE = edges.shape[1]
    nTri = tris.shape[1]

    with ProgressBar(total=nT, ncols=100, dynamic_ncols=False) as pgb:
        dataE, dataB, rows, cols = _matrix_builder(nodes, tets, tris, edges, field.mesh.edge_lengths, tet_to_field, tet_to_edge, ur, er, pgb)
        
    E = coo_matrix((dataE, (rows, cols)), shape=(nE*2 + nTri*2, nE*2 + nTri*2)).tocsr()
    B = coo_matrix((dataB, (rows, cols)), shape=(nE*2 + nTri*2, nE*2 + nTri*2)).tocsr()

    return E, B


############################################################
#           NUMBA ACCELLERATE SUB-MATRIX ASSEMBLY          #
############################################################

@njit(types.Tuple((c16[:,:],c16[:,:]))(f8[:,:], f8[:], i8[:,:], i8[:,:], c16[:,:], c16[:,:]), nogil=True, cache=True, parallel=False, fastmath=True)
def ned2_tet_stiff_mass(tet_vertices, edge_lengths, local_edge_map, local_tri_map, Ms, Mm):
    ''' Nedelec 2 tetrahedral stiffness and mass matrix submatrix Calculation

    '''
    
    Dmat = np.empty((20,20), dtype=np.complex128)
    Fmat = np.empty((20,20), dtype=np.complex128)

    xs, ys, zs = tet_vertices

    bbs, ccs, dds, V = tet_coefficients_bcd(xs, ys, zs)
    b1, b2, b3, b4 = bbs
    c1, c2, c3, c4 = ccs
    d1, d2, d3, d4 = dds
    
    Ds = compute_distances(xs, ys, zs)

    GL1 = np.array([b1, c1, d1]).astype(np.complex128)
    GL2 = np.array([b2, c2, d2]).astype(np.complex128)
    GL3 = np.array([b3, c3, d3]).astype(np.complex128)
    GL4 = np.array([b4, c4, d4]).astype(np.complex128)

    GLs = (GL1, GL2, GL3, GL4)

    letters = [1,2,3,4,5,6]

    KA = 1/(6*V)**4
    KB = 1/(6*V)**2

    V6 = 6*V

    VOLUME_COEFF_CACHE = VOLUME_COEFF_CACHE_BASE*V6
    for ei in range(6):
        ei1, ei2 = local_edge_map[:, ei]
        GA = GLs[ei1]
        GB = GLs[ei2]
        A, B = letters[ei1], letters[ei2]
        L1 = edge_lengths[ei]
        
        
        for ej in range(6):
            ej1, ej2 = local_edge_map[:, ej]
            
            C,D = letters[ej1], letters[ej2]
            
            GC = GLs[ej1]
            GD = GLs[ej2]
            
            VAD = VOLUME_COEFF_CACHE[A,D,0,0]
            VAC = VOLUME_COEFF_CACHE[A,C,0,0]
            VBC = VOLUME_COEFF_CACHE[B,C,0,0]
            VBD = VOLUME_COEFF_CACHE[B,D,0,0]
            VABCD = VOLUME_COEFF_CACHE[A,B,C,D]
            VABCC = VOLUME_COEFF_CACHE[A,B,C,C]
            VABDD = VOLUME_COEFF_CACHE[A,B,D,D]
            VBBCD = VOLUME_COEFF_CACHE[B,B,C,D]
            VABCD = VOLUME_COEFF_CACHE[A,B,C,D]
            VBBCD = VOLUME_COEFF_CACHE[B,B,C,D]
            VAACD = VOLUME_COEFF_CACHE[A,A,C,D]
            VAADD = VOLUME_COEFF_CACHE[A,A,D,D]
            VBBCC = VOLUME_COEFF_CACHE[B,B,C,C]
            VBBDD = VOLUME_COEFF_CACHE[B,B,D,D]
            VAACC = VOLUME_COEFF_CACHE[A,A,C,C]

            L2 = edge_lengths[ej]

            BB1 = matmul(Mm,GC)
            BC1 = matmul(Mm,GD)
            BD1 = dot_c(GA,BB1)
            BE1 = dot_c(GA,BC1)
            BF1 = dot_c(GB,BB1)
            BG1 = dot_c(GB,BC1)

            Q2 = L1*L2
            Q = Q2*9*dot_c(cross_c(GA,GB),matmul(Ms,cross_c(GC,GD)))
            Dmat[ei+0,ej+0] = Q*VAC
            Dmat[ei+0,ej+10] = Q*VAD
            Dmat[ei+10,ej+0] = Q*VBC
            Dmat[ei+10,ej+10] = Q*VBD
            
            Fmat[ei+0,ej+0] = Q2*(VABCD*BD1-VABCC*BE1-VAACD*BF1+VAACC*BG1)
            Fmat[ei+0,ej+10] = Q2*(VABDD*BD1-VABCD*BE1-VAADD*BF1+VAACD*BG1)
            Fmat[ei+10,ej+0] = Q2*(VBBCD*BD1-VBBCC*BE1-VABCD*BF1+VABCC*BG1)
            Fmat[ei+10,ej+10] = Q2*(VBBDD*BD1-VBBCD*BE1-VABDD*BF1+VABCD*BG1)       

        for ej in range(4):
            ej1, ej2, fj = local_tri_map[:, ej]

            C,D,F = letters[ej1], letters[ej2], letters[fj]
            
            GC = GLs[ej1]
            GD = GLs[ej2]
            GF = GLs[fj]

            VABCD = VOLUME_COEFF_CACHE[A,B,C,D]
            VBBCD = VOLUME_COEFF_CACHE[B,B,C,D]
            VAD = VOLUME_COEFF_CACHE[A,D,0,0]
            VAC = VOLUME_COEFF_CACHE[A,C,0,0]
            VAF = VOLUME_COEFF_CACHE[A,F,0,0]
            VBF = VOLUME_COEFF_CACHE[B,F,0,0]
            VBC = VOLUME_COEFF_CACHE[B,C,0,0]
            VBD = VOLUME_COEFF_CACHE[B,D,0,0]
            VABDF = VOLUME_COEFF_CACHE[A,B,D,F]
            VABCF = VOLUME_COEFF_CACHE[A,B,F,C]
            VAADF = VOLUME_COEFF_CACHE[A,A,D,F]
            VAACD = VOLUME_COEFF_CACHE[A,A,C,D]
            VBBDF = VOLUME_COEFF_CACHE[B,B,D,F]
            VBBCD = VOLUME_COEFF_CACHE[B,B,C,D]
            VBBCF = VOLUME_COEFF_CACHE[B,B,F,C]
            VAACF = VOLUME_COEFF_CACHE[A,A,C,F]

            Lab2 = Ds[ej1, ej2]
            Lac2 = Ds[ej1, fj]
            
            AB1 = cross_c(GA,GB)
            AI1 = dot_c(AB1,matmul(Ms,cross_c(GC,GF)))
            AJ1 = dot_c(AB1,matmul(Ms,cross_c(GD,GF)))
            AK1 = dot_c(AB1,matmul(Ms,cross_c(GC,GD)))
            BB1 = matmul(Mm,GF)
            BC1 = matmul(Mm,GC)
            BD1 = matmul(Mm,GD)
            BE1 = dot_c(GA,BB1)
            BF1 = dot_c(GA,BC1)
            BG1 = dot_c(GB,BB1)
            BH1 = dot_c(GB,BC1)
            BI1 = dot_c(GA,BD1)
            BJ1 = dot_c(GB,BD1)
            
            Dmat[ei+0,ej+6] = L1*Lac2*(-6*VAD*AI1-3*VAC*AJ1-3*VAF*AK1)
            Dmat[ei+0,ej+16] = L1*Lab2*(6*VAF*AK1+3*VAD*AI1-3*VAC*AJ1)
            Dmat[ei+10,ej+6] = L1*Lac2*(-6*VBD*AI1-3*VBC*AJ1-3*VBF*AK1)
            Dmat[ei+10,ej+16] = L1*Lab2*(6*VBF*AK1+3*VBD*AI1-3*VBC*AJ1)

            Fmat[ei+0,ej+6] = L1*Lac2*(VABCD*BE1-VABDF*BF1-VAACD*BG1+VAADF*BH1)
            Fmat[ei+0,ej+16] = L1*Lab2*(VABDF*BF1-VABCF*BI1-VAADF*BH1+VAACF*BJ1)
            Fmat[ei+10,ej+6] = L1*Lac2*(VBBCD*BE1-VBBDF*BF1-VABCD*BG1+VABDF*BH1)
            Fmat[ei+10,ej+16] = L1*Lab2*(VBBDF*BF1-VBBCF*BI1-VABDF*BH1+VABCF*BJ1)
    
    ## Mirror the transpose part of the previous iteration as its symmetrical

    Dmat[6:10, :6] = Dmat[:6, 6:10].T
    Fmat[6:10, :6] = Fmat[:6, 6:10].T
    Dmat[16:20, :6] = Dmat[:6, 16:20].T
    Fmat[16:20, :6] = Fmat[:6, 16:20].T
    Dmat[6:10, 10:16] = Dmat[10:16, 6:10].T
    Fmat[6:10, 10:16] = Fmat[10:16, 6:10].T
    Dmat[16:20, 10:16] = Dmat[10:16, 16:20].T
    Fmat[16:20, 10:16] = Fmat[10:16, 16:20].T
    
    for ei in range(4):
        ei1, ei2, fi = local_tri_map[:, ei]
        A, B, E = letters[ei1], letters[ei2], letters[fi]
        GA = GLs[ei1]
        GB = GLs[ei2]
        GE = GLs[fi]
        Lac1 = Ds[ei1, fi]
        Lab1 = Ds[ei1, ei2]
        for ej in range(4):
            ej1, ej2, fj = local_tri_map[:, ej]
            
            C,D,F = letters[ej1], letters[ej2], letters[fj]
            
            GC = GLs[ej1]
            GD = GLs[ej2]
            GF = GLs[fj]

            VABCD = VOLUME_COEFF_CACHE[A,B,C,D]
            VAD = VOLUME_COEFF_CACHE[A,D,0,0]
            VAC = VOLUME_COEFF_CACHE[A,C,0,0]
            VAF = VOLUME_COEFF_CACHE[A,F,0,0]
            VBF = VOLUME_COEFF_CACHE[B,F,0,0]
            VBC = VOLUME_COEFF_CACHE[B,C,0,0]
            VBD = VOLUME_COEFF_CACHE[B,D,0,0]
            VDE = VOLUME_COEFF_CACHE[E,D,0,0]
            VEF = VOLUME_COEFF_CACHE[E,F,0,0]
            VCE = VOLUME_COEFF_CACHE[E,C,0,0]
            VABDF = VOLUME_COEFF_CACHE[A,B,D,F]
            VACEF = VOLUME_COEFF_CACHE[A,C,E,F]
            VABCF = VOLUME_COEFF_CACHE[A,B,F,C]
            VBCDE = VOLUME_COEFF_CACHE[B,C,D,F]
            VBDEF = VOLUME_COEFF_CACHE[B,E,D,F]
            VACDE = VOLUME_COEFF_CACHE[E,A,C,D]
            VBCEF = VOLUME_COEFF_CACHE[B,E,F,C]
            VADEF = VOLUME_COEFF_CACHE[E,A,D,F]

            Lac2 = Ds[ej1, fj]
            Lab2 = Ds[ej1, ej2]

            AB1 = cross_c(GA,GE)
            AF1 = cross_c(GB,GE)
            AG1 = cross_c(GA,GB)
            AH1 = matmul(Ms,cross_c(GC,GF))
            AI1 = matmul(Ms,cross_c(GD,GF))
            AJ1 = matmul(Ms,cross_c(GC,GD))
            AK1 = dot_c(AB1,AH1)
            AL1 = dot_c(AB1,AI1)
            AM1 = dot_c(AB1,AJ1)
            AN1 = dot_c(AF1,AH1)
            AO1 = dot_c(AF1,AI1)
            AP1 = dot_c(AF1,AJ1)
            AQ1 = dot_c(AG1,AH1)
            AR1 = dot_c(AG1,AI1)
            AS1 = dot_c(AG1,AJ1)
            BB1 = matmul(Mm,GF)
            BC1 = matmul(Mm,GC)
            BD1 = matmul(Mm,GD)
            BE1 = dot_c(GE,BB1)
            BF1 = dot_c(GE,BC1)
            BG1 = dot_c(GA,BB1)
            BH1 = dot_c(GA,BC1)
            BI1 = dot_c(GE,BD1)
            BJ1 = dot_c(GA,BD1)
            BK1 = dot_c(GB,BB1)
            BL1 = dot_c(GB,BC1)
            BM1 = dot_c(GB,BD1)

            Q1 = 2*VAD*AN1+VAC*AO1+VAF*AP1
            Q2 = -2*VAF*AP1-VAD*AN1+VAC*AO1
            Dmat[ei+6,ej+6] = Lac1*Lac2*(4*VBD*AK1+2*VBC*AL1+2*VBF*AM1+Q1+2*VDE*AQ1+VCE*AR1+VEF*AS1)
            Dmat[ei+6,ej+16] = Lac1*Lab2*(-4*VBF*AM1-2*VBD*AK1+2*VBC*AL1+Q2-2*VEF*AS1-VDE*AQ1+VCE*AR1)
            Dmat[ei+16,ej+6] = Lab1*Lac2*(-4*VDE*AQ1-2*VCE*AR1-2*VEF*AS1-2*VBD*AK1-VBC*AL1-VBF*AM1+Q1)
            Dmat[ei+16,ej+16] = Lab1*Lab2*(4*VEF*AS1+2*VDE*AQ1-2*VCE*AR1+2*VBF*AM1+VBD*AK1-VBC*AL1+Q2)
            Fmat[ei+6,ej+6] = Lac1*Lac2*(VABCD*BE1-VABDF*BF1-VBCDE*BG1+VBDEF*BH1)
            Fmat[ei+6,ej+16] = Lac1*Lab2*(VABDF*BF1-VABCF*BI1-VBDEF*BH1+VBCEF*BJ1)
            Fmat[ei+16,ej+6] = Lab1*Lac2*(VBCDE*BG1-VBDEF*BH1-VACDE*BK1+VADEF*BL1)
            Fmat[ei+16,ej+16] = Lab1*Lab2*(VBDEF*BH1-VBCEF*BJ1-VADEF*BL1+VACEF*BM1)

    Dmat = Dmat*KA
    Fmat = Fmat*KB

    return Dmat, Fmat


############################################################
#             NUMBA ACCELLERATED MATRIX BUILDER            #
############################################################

@njit(types.Tuple((c16[:], c16[:], i8[:], i8[:]))(f8[:,:], 
                                                      i8[:,:], 
                                                      i8[:,:], 
                                                      i8[:,:], 
                                                      f8[:], 
                                                      i8[:,:], 
                                                      i8[:,:], 
                                                      c16[:,:,:], 
                                                      c16[:,:,:], 
                                                      ProgressBarType), cache=True, nogil=True, parallel=True)
def _matrix_builder(nodes, tets, tris, edges, all_edge_lengths, tet_to_field, tet_to_edge, ur, er, pgb: ProgressBar):
    nT = tets.shape[1]
    nedges = edges.shape[1]

    nnz = nT*400

    rows = np.empty(nnz, dtype=np.int64)
    cols = np.empty_like(rows)
    dataE = np.empty_like(rows, dtype=np.complex128)
    dataB = np.empty_like(rows, dtype=np.complex128)

    
    for itet in prange(nT): # ty: ignore
        p = itet*400
        if np.mod(itet,10)==0:
            pgb.update(10)
        urt = ur[:,:,itet]
        ert = er[:,:,itet]

        # Construct a local mapping to global triangle orientations
        
        local_tri_map = local_tet_to_triid(tet_to_field, tets, tris, itet, nedges)
        local_edge_map = local_tet_to_edgeid(tets, edges, tet_to_field, itet)
        edge_lengths = all_edge_lengths[tet_to_edge[:,itet]]

        # Construct the local edge map

        Esub, Bsub = ned2_tet_stiff_mass(nodes[:,tets[:,itet]], 
                                                edge_lengths, 
                                                local_edge_map, 
                                                local_tri_map, 
                                                matinv(urt), ert)
        
        indices = tet_to_field[:, itet]
        for ii in range(20):
            rows[p+20*ii:p+20*(ii+1)] = indices[ii]
            cols[p+ii:p+400+ii:20] = indices[ii]

        dataE[p:p+400] = Esub.ravel()
        dataB[p:p+400] = Bsub.ravel()
    return dataE, dataB, rows, cols





# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

from __future__ import annotations
import numpy as np
from ..mesh3d import SurfaceMesh
from .femdata import FEMBasis
from .ned2_interp import ned2_tri_interp_full, ned2_tri_interp_curl
from ..mth.optimized import matinv
from ..cs import CoordinateSystem
from ..const import MU0, C0

## TODO: TEMPORARY SOLUTION FIX THIS

class FieldFunctionClass:
    """ This Class serves as a picklable class so that ModalPort boundary conditions
    can actually be stored with the Simulation3D class. Functions aren't picklable in 
    Python. 
    
    I am not happy with the existence of this class, it feels too ad-hoc but for now it
    is the simplest way. It stores all actually required information needed to do a
    surface field interpolation without needing to store the Mesh3D and SurfaceMesh class
    objects plus the NedelecLegrange2 classes with the Simulation3D. 

    As it stands currently, only the GMSH mesh is stored plus the geometry objects. The
    mesh is reconstructed as it is deterministic.
    """
    def __init__(self, 
                 field: np.ndarray,
                 cs: CoordinateSystem,
                 nodes: np.ndarray,
                 tris: np.ndarray,
                 tri_to_field: np.ndarray,
                 EH: str = 'E',
                 diadic: np.ndarray | None = None,
                 beta: float | None = None,
                 constant: float | int | complex = 1.0):
        self.field: np.ndarray = field
        self.cs: CoordinateSystem = cs
        self.nodes: np.ndarray = nodes
        self.tris: np.ndarray = tris
        self.tri_to_field: np.ndarray = tri_to_field
        self.eh: str = EH
        self.diadic: np.ndarray | None = diadic
        self.beta: float | None = beta
        self.constant: float = constant
        if EH == 'H':
            if diadic is None:
                self.diadic = np.eye(3)[:,:,np.newaxis()] * np.ones((self.tris.shape[1])) # type: ignore

    def __call__(self, xs: np.ndarray,
             ys: np.ndarray,
             zs: np.ndarray):
        xl, yl, zl = self.cs.in_local_cs(xs, ys, zs)
        if self.eh == 'E':
            Fxl, Fyl, Fzl = self.calcE(xl, yl)
        else:
            Fxl, Fyl, Fzl = self.calcH(xl, yl)
        Fx, Fy, Fz = self.cs.in_global_basis(Fxl, Fyl, Fzl)
        return np.array([Fx, Fy, Fz])*self.constant
    
    def calcE(self, xs: np.ndarray, ys: np.ndarray) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        coordinates = np.array([xs, ys])
        return ned2_tri_interp_full(coordinates, 
                               self.field, 
                               self.tris,  
                               self.nodes, 
                               self.tri_to_field)
    
    def calcH(self, xs: np.ndarray, ys: np.ndarray) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        coordinates = np.array([xs, ys])
        
        return ned2_tri_interp_curl(coordinates, 
                               self.field, 
                               self.tris,  
                               self.nodes, 
                               self.tri_to_field,
                               self.diadic,
                               self.beta)

############### Nedelec2 Class

class NedelecLegrange2(FEMBasis):


    def __init__(self, mesh: SurfaceMesh, cs: CoordinateSystem):

        self.mesh: SurfaceMesh = mesh

        self.cs: CoordinateSystem = cs

        ## 
        nodes = self.mesh.nodes
        self.local_nodes: np.ndarray = np.array(self.cs.in_local_cs(nodes[0,:], nodes[1,:], nodes[2,:]))
                                                
        ## Counters
        self.n_nodes: int = self.mesh.n_nodes
        self.n_edges: int = self.mesh.n_edges
        self.n_tris: int = self.mesh.n_tris
        self.n_tri_dofs: int = None

        self.n_field: int = 2*self.n_edges + 2*self.n_tris + self.n_nodes + self.n_edges
        self.n_xy: int = 2*self.n_edges + 2*self.n_tris

        ######## MESH Derived
        Nn = self.mesh.n_nodes
        Ne = self.mesh.n_edges
        Nt = self.mesh.n_tris

        self.tri_to_field: np.ndarray = np.zeros((8 + 6, self.n_tris), dtype=int)

        self.tri_to_field[:3,:] = self.mesh.tri_to_edge
        self.tri_to_field[3,:] = np.arange(Nt) + Ne
        self.tri_to_field[4:7,:] = self.mesh.tri_to_edge + Ne + Nt
        self.tri_to_field[7,:] = np.arange(Nt) + 2*Ne + Nt
        self.tri_to_field[8:11,:] = self.mesh.tris + (2*Ne + 2*Nt) # + E + T + E + T
        self.tri_to_field[11:14,:] = self.mesh.tri_to_edge + (2*Ne + 2*Nt + Nn)
    
        self.edge_to_field: np.ndarray = np.zeros((5,Ne), dtype=int) #edge mode 1, edge mode 2, edge legrande mode, edge vertex mode 1, edge vertex mode 2

        self.edge_to_field[0,:] = np.arange(Ne)
        self.edge_to_field[1,:] = np.arange(Ne) + Nt + Ne
        self.edge_to_field[2,:] = np.arange(Ne) + Ne*2 + Nt*2 + Nn
        self.edge_to_field[3:,:] = self.mesh.edges + Ne*2 + Nt*2

        ##
        self._field: np.ndarray = None   
        self._rows: np.ndarray = None
        self._cols: np.ndarray = None 

    def __call__(self, **kwargs) -> NedelecLegrange2:
        self._field = self.fielddata(**kwargs)
        return self
    
    def interpolate_Ef(self, field: np.ndarray) -> FieldFunctionClass:
        '''Generates the Interpolation function as a function object for a given coordiante basis and origin.'''
        
        # def func(xs: np.ndarray, ys: np.ndarray, zs: np.ndarray) -> np.ndarray:
        #     xl, yl, zl = self.cs.in_local_cs(xs, ys, zs)
        #     Exl, Eyl, Ezl = self.tri_interpolate(field, xl, yl)
        #     Ex, Ey, Ez = self.cs.in_global_basis(Exl, Eyl, Ezl)
        #     return np.array([Ex, Ey, Ez])
        return FieldFunctionClass(field, self.cs, self.local_nodes, self.mesh.tris, self.tri_to_field, 'E')

    def interpolate_Hf(self, field: np.ndarray, k0: float, ur: np.ndarray, beta: float) -> FieldFunctionClass:
        '''Generates the Interpolation function as a function object for a given coordiante basis and origin.'''
        constant = 1j / ((k0*C0)*MU0)
        urinv = np.zeros_like(ur)
        
        for i in range(ur.shape[2]):
            urinv[:,:,i] = matinv(ur[:,:,i])

        # def func(xs: np.ndarray, ys: np.ndarray, zs: np.ndarray) -> np.ndarray:
        #     xl, yl, _ = self.cs.in_local_cs(xs, ys, zs)
        #     Exl, Eyl, Ezl = self.tri_interpolate_curl(field, xl, yl, urinv, beta)
        #     Ex, Ey, Ez = self.cs.in_global_basis(Exl, Eyl, Ezl)
        #     return np.array([Ex, Ey, Ez])*constant
        return FieldFunctionClass(field, self.cs, self.local_nodes, self.mesh.tris, self.tri_to_field, 'H', urinv, beta, constant)
    
    def tri_interpolate(self, field, xs: np.ndarray, ys: np.ndarray) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        coordinates = np.array([xs, ys])
        return ned2_tri_interp_full(coordinates, 
                               field, 
                               self.mesh.tris,  
                               self.local_nodes, 
                               self.tri_to_field)
    
    def tri_interpolate_curl(self, field, xs: np.ndarray, ys: np.ndarray, diadic: np.ndarray | None = None, beta: float = 0.0) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        coordinates = np.array([xs, ys])
        if diadic is None:
            diadic = np.eye(3)[:,:,np.newaxis()] * np.ones((self.mesh.n_tris)) # type: ignore
        return ned2_tri_interp_curl(coordinates, 
                               field, 
                               self.mesh.tris,  
                               self.local_nodes, 
                               self.tri_to_field,
                               diadic,
                               beta)
    
    
    # def interpolate_curl(self, field: np.ndarray, xs: np.ndarray, ys: np.ndarray, zs:np.ndarray, c: np.ndarray) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    #     """
    #     Interpolates the curl of the field at the given points.
    #     """
    #     return ned2_tet_interp_curl(np.array([xs, ys,zs]), field, self.mesh.tets, self.mesh.tris, self.mesh.edges, self.mesh.nodes, self.tet_to_field, c)
    
    # def fieldf(self, field: np.ndarray, basis: np.ndarray = None, origin: np.ndarray = None) -> Callable:
    #     if basis is None:
    #         basis = np.eye(3)

    #     if origin is None:
    #         origin = np.zeros(3)
        
    #     ibasis = np.linalg.pinv(basis)
    #     def func(xs: np.ndarray, ys: np.ndarray, zs: np.ndarray) -> np.ndarray:
    #         xyz = np.array([xs, ys, zs]) + origin[:, np.newaxis]
    #         xyzg = basis @ xyz
    #         return ibasis @ np.array(self.interpolate(field, xyzg[0,:], xyzg[1,:], xyzg[2,:]))
    #     return func
    
    ###### INDEX MAPPINGS

    